require 'json'


#===============================================================================
# FXF
#
module FXF
	# version 1.0
	VERSION = '1.0'
	
	#---------------------------------------------------------------------------
	# generate
	#
	
	# Generates a FXF JSON string.
	
	def self.generate(struct, opts={})
		# $tm.hrm
		rv = {}
		self.generate_object(rv, struct)
		rv = {'root'=>rv.keys[0], 'objects'=>rv}
		
		# return JSON
		if opts['pretty']
			return JSON.pretty_generate(rv)
		else
			return JSON.generate(rv)
		end
	end
	#
	# generate
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# parse
	#
	
	# Parses an FXF string and returns the structure it defines.
	
	def self.parse(fxf)
		parser = FXF::Parser.new(fxf)
		return parser.parse()
	end
	#
	# parse
	#---------------------------------------------------------------------------
	
	
	# private
	private
	
	
	#---------------------------------------------------------------------------
	# generate_object
	#
	
	# Scalar object types. These are all the data types in JSON except for
	# hashes and arrays.
	SCALARS = [String, Integer, Float, TrueClass, FalseClass, NilClass]
	
	def self.generate_object(rv, obj)
		# $tm.hrm
		key = obj.object_id.to_s
		
		# add to rv
		if not rv.has_key?(key)
			# scalar
			if SCALARS.include?(obj.class)
				rv[key] = obj
				
			# hash
			elsif obj.is_a?(Hash)
				hsh = rv[key] = {}
				
				obj.each do |k, v|
					hsh[k] = self.generate_object(rv, v)
				end
				
			# array
			elsif obj.is_a?(Array)
				arr = rv[key] = []
				
				obj.each do |v|
					arr.push self.generate_object(rv, v)
				end
				
			# to_fxf
			elsif obj.respond_to?('to_fxf')
				dfn = {}
				dfn['class'] = obj.class.to_s
				dfn['details'] = obj.to_fxf()
				object_dfn rv, key, 'custom', dfn
				
			# standard class
			elsif translator = FXF::Standard::TO_FXF[obj.class.to_s]
				dfn = translator.call(obj)
				object_dfn rv, key, 'standard', dfn
			
			# else unknown class
			else
				# build details
				details = {}
				
				# string
				if obj.respond_to?('to_s')
					details['str'] = obj.to_s
				end
				
				# buid definition
				dfn = {}
				dfn['scope'] = 'unrecognized'
				dfn['class'] = obj.class.to_s
				dfn['details'] = details
				
				# store
				rv[key] = [dfn]
			end
		end
		
		# return
		return key
	end
	#
	# generate_object
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# object_dfn
	#
	def self.object_dfn(rv, key, scope, dfn)
		# $tm.hrm
		export = {}
		rv[key] = [export]
		export['scope'] = scope
		export['class'] = dfn['class']
		export['details'] = {}
		
		# add object references
		if dfn['details']
			if dfn['details'].is_a?(Hash)
				dfn['details'].each do |k, v|
					export['details'][k] = self.generate_object(rv, v)
				end
			else
				raise 'custom-object-details-must-be-hash'
			end
		end
	end
	#
	# object_dfn
	#---------------------------------------------------------------------------
end
#
# FXF
#===============================================================================


#===============================================================================
# FXF::Standard
#

# This module defines procs for serializing and deserializing DateTime and
# URI::HTTPS objects.

module FXF::Standard
	TO_FXF = {}
	FROM_FXF = {}
	
	#---------------------------------------------------------------------------
	# DateTime
	#
	TO_FXF['DateTime'] = proc do |obj|
		dfn = {}
		dfn['class'] = 'datetime'
		dfn['details'] = {'str'=>obj.to_s}
		dfn
	end
	
	FROM_FXF['datetime'] = proc do |dfn|
		# $tm.hrm
		DateTime.parse(dfn.details['str'])
	end
	#
	# DateTime
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# URI
	#
	TO_FXF['URI::HTTPS'] = proc do |obj|
		dfn = {}
		dfn['class'] = 'uri'
		dfn['details'] = {'str'=>obj.to_s}
		dfn
	end

	FROM_FXF['uri'] = proc do |dfn|
		URI(dfn.details['str'])
	end
	#
	# URI
	#---------------------------------------------------------------------------
end
#
# FXF::Standard
#===============================================================================


#===============================================================================
# FXF::Parser
#

# `FXF::Parser` isn't intended to be used directly. FXF.parse creates a
# `FXF::Parser` and calls its `parse` method.

class FXF::Parser
	#---------------------------------------------------------------------------
	# initialize
	#
	def initialize(raw)
		@raw = raw
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# parse
	#
	def parse()
		@full = JSON.parse(@raw)
		@org = @full['objects']
		@found = {}
		@collections = []
		rv = self.build_object(@full['root'])
		
		# look for object place holders
		placeholder_mod()
		
		# return
		return rv
	end
	#
	# parse
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# placeholder_mod
	#
	def placeholder_mod
		# $tm.hrm
		placeholders = []
		phs_to_obj = {}
		
		# build unique list of placeholders
		@collections.each do |collection|
			if collection.is_a?(Array)
				collection.each do |el|
					placeholder_add placeholders, el
				end
			elsif collection.is_a?(Hash)
				collection.values.each do |el|
					placeholder_add placeholders, el
				end
			end
		end
		
		# build a hash of placeholders to objects
		placeholders.each do |ph|
			# standard
			if ph.scope == 'standard'
				if translator = FXF::Standard::FROM_FXF[ph.clss]
					phs_to_obj[ph] = translator.call(ph)
				else
					raise 'do-not-have-standard-translator: ' + ph.clss
				end
			
			# custom
			elsif ph.scope == 'custom'
				if Module.const_defined?(ph.clss)
					clss = Module.const_get(ph.clss)
					
					if clss.respond_to?('from_fxf')
						phs_to_obj[ph] = clss.from_fxf(ph.details)
					else
						raise 'class-does-not-have-from-fxf: ' + ph.clss
					end
				else
					raise 'unknown-custom-class: ' + ph.clss
				end
			
			# else don't know how to build this object
			else
				phs_to_obj[ph] = ph
			end
		end
		
		# substitute placeholders to objects
		@collections.each do |collection|
			if collection.is_a?(Array)
				collection.map! do |el|
					rv = nil
					
					if phs_to_obj.has_key?(el)
						rv = phs_to_obj[el]
					else
						rv = el
					end
					
					rv
				end
			
			elsif collection.is_a?(Hash)
				collection.each do |k, v|
					if phs_to_obj.has_key?(v)
						collection[k] = phs_to_obj[v]
					end
				end
			end
		end
	end
	#
	# placeholder_mod
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# placeholder_add
	#
	def placeholder_add(placeholders, el)
		# $tm.hrm
		
		# if placeholder, and not in placeholders array, add to array
		if el.is_a?(FXF::Parser::ObjectHolder) and (not placeholders.include?(el))
			placeholders.push el
		end
	end
	#
	# placeholder_add
	#---------------------------------------------------------------------------


	#---------------------------------------------------------------------------
	# build_object
	#
	def build_object(key)
		# $tm.hrm
		rv = nil
		
		# if already found
		if @found.has_key?(key)
			return @found[key]
		end
		
		# get object definition
		dfn = @org.delete(key)
		
		# scalar
		if FXF::SCALARS.include?(dfn.class)
			@found[key] = dfn
			return dfn
			
		# array
		elsif dfn.is_a?(Array)
			# if object description
			if dfn[0].is_a?(Hash)
				details = {}
				build_hash(dfn[0]['details'], details)
				dfn = FXF::Parser::ObjectHolder.new dfn[0]['scope'], dfn[0]['class'], details
				@found[key] = dfn
				return dfn
			else
				@found[key] = []
				@collections.push @found[key]
				return build_array(dfn, @found[key])
			end
			
		# hash
		elsif dfn.is_a?(Hash)
			@found[key] = {}
			@collections.push @found[key]
			return build_hash(dfn, @found[key])
			
		# else unknown
		else
			puts 'unknown: ' + dfn.class.to_s
			exit
		end
		
		# return
		return rv
	end
	#
	# build_object
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# build_hash
	#
	def build_hash(dfn, hsh)
		# loop through elements
		dfn.each do |key, ref|
			hsh[key] = build_object(ref)
		end
		
		# return
		return hsh
	end
	#
	# build_hash
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# build_array
	#
	def build_array(dfn, arr)
		# loop through elements
		dfn.each do |ref|
			arr.push build_object(ref)
		end
		
		# return
		return arr
	end
	#
	# build_array
	#---------------------------------------------------------------------------
end
#
# FXF::Parser
#===============================================================================


#===============================================================================
# FXF::Parser::ObjectHolder
#

# Objects of this class are placeholders for objects that have been serialized.
# This class is used by FXF::Parser and is not meant to be used directly.

class FXF::Parser::ObjectHolder
	attr_reader :scope
	attr_reader :clss
	attr_reader :details
	
	#---------------------------------------------------------------------------
	# initialize
	#
	def initialize(scope, clss, details)
		@scope = scope
		@clss = clss
		@details = details
	end
	#
	# initialize
	#---------------------------------------------------------------------------
end
#
# FXF::Parser::ObjectHolder
#===============================================================================